/*  generic.c

    Generic file for  krenzo  (X11 contour field tool for Karma).

    Copyright (C) 1996-2000  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable interactive viewing of a contour field
    (otherwise known as a "Renzogram", after Renzo Sanchissi).
    This module runs on an X11 server.


    Written by      Richard Gooch   4-SEP-1996: Copied from kpvslice module.

    Updated by      Richard Gooch   6-SEP-1996: Created <change_active_states>.

    Updated by      Richard Gooch   12-SEP-1996: Move out call to
  <change_active_states>.

    Updated by      Richard Gooch   14-SEP-1996: Removed moment computation.

    Updated by      Richard Gooch   28-OCT-1996: Added hostname and port number
  to title.

    Updated by      Richard Gooch   27-DEC-1996: Switched to
  <iarray_create_and_setup>.

    Updated by      Richard Gooch   31-DEC-1996: Now return min&max from
  <process_image>.

    Updated by      Richard Gooch   2-FEB-1996: No longer destroy KwcsAstro
  objects.

    Updated by      Richard Gooch   5-SEP-1998: Removed comms. setup code.

    Updated by      Richard Gooch   16-SEP-1998: Made use of new key grouping
  in <contour> package.

    Last updated by Richard Gooch   12-JUL-2000: Added
  #include <karma_canvas.h>.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <errno.h>
#include <X11/Xlib.h>
#include <karma.h>
#include <karma_viewimg.h>
#include <karma_contour.h>
#include <karma_canvas.h>
#include <karma_iarray.h>
#include <karma_conn.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_r.h>
#include "krenzo.h"


/*  Private functions  */
STATIC_FUNCTION (void destroy_all_vimages,
		 (ViewableImage *image, ViewableImage *magnified_image) );
STATIC_FUNCTION (void destroy_all_cimages, () );


/*  Private data  */


/*  Public functions follow  */

iarray process_image (multi_array *multi_desc, double *min, double *max,
		      KWorldCanvas pseudo_canvas,
		      KWorldCanvas mag_pseudo_canvas,
		      ViewableImage *image, ViewableImage *magnified_image)
/*  [PURPOSE] This routine will load a file and display it.
    <multi_desc> The multi_array descriptor.
    <min> The minimum value will be written here.
    <max> The maximum value will be written here.
    <pseudo_canvas> The PseudoColour canvas.
    <image> If an image is loaded the ViewableImage is written here. If no
    image is loaded, NULL is written here. The value written here must be
    preserved between calls.
    <magnified_image> If an image is loaded the magnified ViewableImage is
    written here. If no image is loaded, NULL is written here. The value
    written here must be preserved between calls.
    [RETURNS] The image array on success, else NULL.
*/
{
    iarray image_pseudo = NULL;
    extern KwcsAstro loaded_image_ap, main_ap;
    extern unsigned int image_mode;
    /*static char function_name[] = "process_image";*/

    if ( !iarray_create_and_setup (&image_pseudo, multi_desc, TRUE, 2, K_FLOAT,
				   min, max, TRUE, &loaded_image_ap) )
    {
	return (NULL);
    }
    if (image_mode == IMAGE_MODE_LOADED) main_ap = NULL;
    destroy_all_vimages (image, magnified_image);
    if ( ( *image = viewimg_create_from_iarray (pseudo_canvas, image_pseudo,
						FALSE) ) == NULL )
    {
	fprintf (stderr, "Error getting ViewableImage from Iarray\n");
	iarray_dealloc (image_pseudo);
	loaded_image_ap = NULL;
	return (NULL);
    }
    if ( ( *magnified_image =
	   viewimg_create_from_iarray (mag_pseudo_canvas, image_pseudo,
				       FALSE) ) == NULL )
    {
	fprintf (stderr, "Error getting ViewableImage from Iarray\n");
	iarray_dealloc (image_pseudo);
	loaded_image_ap = NULL;
	destroy_all_vimages (image, magnified_image);
	return (NULL);
    }
    canvas_set_attributes (pseudo_canvas,
			   CANVAS_ATT_VALUE_MIN, *min,
			   CANVAS_ATT_VALUE_MAX, *max,
			   CANVAS_ATT_END);
    canvas_set_attributes (mag_pseudo_canvas,
			   CANVAS_ATT_VALUE_MIN, *min,
			   CANVAS_ATT_VALUE_MAX, *max,
			   CANVAS_ATT_END);
    if (image_mode == IMAGE_MODE_LOADED)
    {
	if ( !viewimg_make_active (*image) )
	{
	    fprintf (stderr, "Error making ViewableImage active\n");
	    iarray_dealloc (image_pseudo);
	    destroy_all_vimages (image, magnified_image);
	    loaded_image_ap = NULL;
	    return (NULL);
	}
	if ( !viewimg_make_active (*magnified_image) )
	{
	    fprintf (stderr, "Error making ViewableImage active\n");
	    iarray_dealloc (image_pseudo);
	    destroy_all_vimages (image, magnified_image);
	    loaded_image_ap = NULL;
	    return (NULL);
	}
	main_ap = loaded_image_ap;
    }
    return (image_pseudo);
}   /*  End Function process_image  */

iarray process_cube (multi_array *multi_desc, double *min, double *max)
/*  [PURPOSE] This routine will load a cube.
    <multi_desc> The multi_array descriptor.
    <min> The minimum value will be written here.
    <max> The maximum value will be written here.
    [RETURNS] The cube array on success, else NULL.
*/
{
    iarray cube = NULL;
    extern KWorldCanvas main_canvas;
    extern KwcsAstro cube_ap;
    extern unsigned int num_cimages;
    extern KContourImage *cimages;
    /*static char function_name[] = "load_cube";*/

    if ( !iarray_create_and_setup (&cube, multi_desc, TRUE, 3, K_FLOAT,
				   min, max, TRUE, &cube_ap) )
    {
	return (NULL);
    }
    destroy_all_cimages ();
    if ( ( cimages = contour_create_sequence_from_iarray (main_canvas, cube,
							  2, 1, 0, 0, NULL,0) )
	 == NULL )
    {
	fprintf (stderr, "Error creating contour images\n");
	iarray_dealloc (cube);
	cube_ap = NULL;
	return (NULL);
    }
    num_cimages = iarray_dim_length (cube, 0);
    return (cube);
}   /*  End Function process_cube  */

void change_active_states ()
/*  [SUMMARY] Set active states for all KContourImages.
    [RETURNS] Nothing.
*/
{
    int count;
    extern flag change_active_cimages;
    extern int start_channel, end_channel, channel_skip;
    extern unsigned int num_cimages;
    extern KContourImage *cimages;

    /*  First make everyone inactive  */
    for (count = 0; count < num_cimages; ++count)
    {
	contour_set_active (cimages[count], FALSE, FALSE, FALSE, FALSE, 0);
    }
    /*  Now make desired ones active  */
    for (count = start_channel; count <= end_channel; count += channel_skip)
    {
	contour_set_active (cimages[count], TRUE, FALSE, FALSE, FALSE, 0);
    }
    change_active_cimages = FALSE;
}   /*  End Function change_active_states  */

void convert_lin_world_coords (double *xout, double *yout,
			       KwcsAstro out_ap,
			       CONST double *xin, CONST double *yin,
			       KwcsAstro in_ap, unsigned int num_coords)
/*  [SUMMARY] Convert linear world co-ordinates in two reference frames.
    [PURPOSE] This routine will convert linear world co-ordinates in one
    reference frame to linear world co-ordinates in another frame. If
    necessary, linear to non-linear and non-linear to linear conversions are
    performed.
    <xout> The output horizontal co-ordinates are written here.
    <yout> The output vertical co-ordinates are written here.
    <out_ap> The output reference frame.
    <xin> The input horizontal co-ordinates are written here.
    <yin> The input vertical co-ordinates are written here.
    <in_ap> The input reference frame.
    [RETURNS] Nothing.
*/
{
    unsigned int count;

    /*  First copy to the output arrays  */
    for (count = 0; count < num_coords; ++count)
    {
	xout[count] = xin[count];
	yout[count] = yin[count];
    }
    if ( !wcs_astro_test_radec (in_ap) && !wcs_astro_test_radec (out_ap) )
    {
	/*  Both reference frames are linear: plain copy  */
	return;
    }
    if ( wcs_astro_test_radec (in_ap) && wcs_astro_test_radec (out_ap) )
    {
	/*  Both reference frames are non-linear: convert back and forth  */
	wcs_astro_transform (in_ap, num_coords,
			     xout, FALSE, yout, FALSE, NULL, FALSE,
			     0, NULL, NULL);
	wcs_astro_transform (out_ap, num_coords,
			     xout, TRUE, yout, TRUE, NULL, FALSE,
			     0, NULL, NULL);
	return;
    }
    if ( wcs_astro_test_radec (in_ap) )
    {
	/*  Input reference frame is non-linear but output is linear: convert
	    back to linear  */
	wcs_astro_transform (in_ap, num_coords,
			     xout, TRUE, yout, TRUE, NULL, FALSE,
			     0, NULL, NULL);
	return;
    }
    /*  Output reference frame is non-linear but input is linear: do nothing */
}   /*  End Function convert_lin_world_coords  */


/*  Private functions follow  */

static void destroy_all_vimages (ViewableImage *image,
				 ViewableImage *magnified_image)
/*  [PURPOSE] This routine will destroy all ViewableImages and clear the
    appropriate pointers.
    [RETURNS] Nothing.
*/
{
    if (*image != NULL) viewimg_destroy (*image);
    *image = NULL;
    if (*magnified_image != NULL) viewimg_destroy (*magnified_image);
    *magnified_image = NULL;
}   /*  End Function destroy_all_vimages  */

static void destroy_all_cimages ()
/*  [SUMMARY] Destroy all contourable images.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    extern unsigned int num_cimages;
    extern KContourImage *cimages;

    for (count = 0; count < num_cimages; ++count)
    {
	contour_destroy (cimages[count]);
	cimages[count] = NULL;
    }
    num_cimages = 0;
}   /*  End Function destroy_all_cimages  */


/*  Private routines follow  */
